import json
import traceback
import typing

from django.conf import settings
from django.http import HttpRequest, HttpResponse
from django.utils.translation import ugettext_lazy as _
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from gql.playground import PLAYGROUND_HTML
from gql.utils import place_files_in_operations
from graphql import ExecutionResult, GraphQLError, GraphQLSchema, graphql_sync

from .exceptions import GraphQLExtensionError, MethodNotAllowedError, UserInputError
from .response import Response
from .settings import api_settings


class GraphQLView(View):
    schema: GraphQLSchema

    context_builder: typing.Callable = None

    def __init__(self, **kwargs):
        self.schema = api_settings.SCHEMA
        self.context_builder = api_settings.CONTEXT_BUILDER
        super().__init__(**kwargs)

    @classmethod
    def as_view(cls, **initkwargs):
        view = super().as_view(**initkwargs)
        return csrf_exempt(view)

    def dispatch(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        try:
            method = request.method.lower()
            if method == 'get' and api_settings.ENABLE_PLAYGROUND:
                return HttpResponse(PLAYGROUND_HTML)

            if method != 'post':
                raise MethodNotAllowedError()

            data = self.parse_body(request)
            return self.get_response(request, data)
        except GraphQLExtensionError as e:
            return Response(data={'errors': [e.formatted]})

    def format_error(self, error: GraphQLError):
        if not error:
            raise ValueError("Received null or undefined error.")
        formatted = dict(  # noqa: E701 (pycqa/flake8#394)
            message=error.message or "An unknown error occurred.",
            locations=[loc._asdict() for loc in error.locations] if error.locations else None,
            path=error.path,
        )
        if settings.DEBUG and error.original_error:
            original_error = error.original_error
            exception = error.extensions.get('exception', {})
            exception['traceback'] = traceback.format_exception(
                type(original_error), original_error, original_error.__traceback__
            )
            error.extensions['exception'] = exception
        if error.extensions:
            formatted.update(extensions=error.extensions)
        return formatted

    def get_response(self, request: HttpRequest, data: dict) -> Response:
        query, variables, operation_name, id = self.get_graphql_params(request, data)

        execution_result = self.execute_graphql_request(request, query, variables, operation_name)

        data = {}
        if not execution_result:
            return Response(data)

        if execution_result.errors:
            data['errors'] = [self.format_error(e) for e in execution_result.errors]
        data['data'] = execution_result.data
        return Response(data)

    def parse_body(self, request: HttpRequest) -> dict:
        content_type = self.get_content_type(request)

        if content_type == 'application/graphql':
            return {'query': request.body.decode()}

        elif content_type == 'application/json':
            try:
                body = request.body.decode()
            except Exception as e:
                raise UserInputError(str(e))

            try:
                return json.loads(body)
            except (TypeError, ValueError):
                raise UserInputError(_('POST body sent invalid JSON.'))

        elif content_type == 'multipart/form-data':
            body = request.POST
            try:
                operations = json.loads(body.get('operations', '{}'))
                files_map = json.loads(body.get('map', '{}'))
            except (TypeError, ValueError):
                raise UserInputError(_('operations or map sent invalid JSON.'))
            if not files_map and not operations:
                return body
            return place_files_in_operations(operations, files_map, request.FILES)

        elif content_type == 'application/x-www-form-urlencoded':
            return request.POST

        return {}

    def execute_graphql_request(self, request, query, variables, operation_name) -> ExecutionResult:
        if not query:
            raise UserInputError(_('Must provide query string.'))
        if self.context_builder:
            context = self.context_builder()
        else:
            context = {}
        context['request'] = request
        return graphql_sync(
            self.schema, query, variable_values=variables, context_value=context, operation_name=operation_name
        )

    @staticmethod
    def json_encode(d):
        return json.dumps(d, separators=(',', ':'))

    @staticmethod
    def get_graphql_params(request, data):
        query = request.GET.get('query') or data.get('query')
        variables = request.GET.get('variables') or data.get('variables')
        id = request.GET.get('id') or data.get('id')

        if variables and isinstance(variables, str):
            try:
                variables = json.loads(variables)
            except Exception:
                raise UserInputError(_('Variables are invalid JSON.'))

        operation_name = request.GET.get('operationName') or data.get('operationName')
        if operation_name == 'null':
            operation_name = None

        return query, variables, operation_name, id

    @staticmethod
    def get_content_type(request):
        meta = request.META
        content_type = meta.get('CONTENT_TYPE', meta.get('HTTP_CONTENT_TYPE', ''))
        return content_type.split(';', 1)[0].lower()



from functools import wraps

from djgql.exceptions import AuthenticationError


def login_required(func):
    @wraps(func)
    def wrap(parent, info, *args, **kwargs):
        request = info.context['request']
        if 'authenticated' not in request.auth.scopes:
            raise AuthenticationError()
        info.context['user'] = request.user

        return func(*args, **kwargs)

    return wrap



from django.http.response import JsonResponse


class Response(JsonResponse):
    def __init__(self, data, **kwargs):
        json_dumps_params = kwargs.get('json_dumps_params', {})
        json_dumps_params['separators'] = (',', ':')
        kwargs['json_dumps_params'] = json_dumps_params
        super().__init__(data, **kwargs)



"""
Provides various authentication policies.
"""
import base64
import binascii
import typing

from django.contrib.auth import authenticate, get_user_model
from django.contrib.auth.models import AnonymousUser
from django.utils.translation import ugettext_lazy as _

from djgql.exceptions import AuthenticationError


class AuthCredentials:
    def __init__(self, scopes: typing.Sequence[str] = None):
        self.scopes = [] if scopes is None else list(scopes)


class UnauthenticatedUser(AnonymousUser):
    def __init__(self, error: AuthenticationError = None):
        self.error = error


def get_authorization_header(request):
    """
    Return request's 'Authorization:' header, as a bytestring.

    Hide some test client ickyness where the header can be unicode.
    """
    auth = request.META.get('HTTP_AUTHORIZATION', b'')
    if isinstance(auth, str):
        auth = auth.encode()
    return auth


class BaseAuthMiddleware:
    def __init__(self, get_response) -> None:
        self.get_response = get_response

    def authenticate(self, request):
        """
        Authenticate the request and return a two-tuple of (user, token).
        """
        raise NotImplementedError(".authenticate() must be overridden.")

    def authenticate_header(self, request):
        """
        Return a string to be used as the value of the `WWW-Authenticate`
        header in a `401 Unauthenticated` response, or `None` if the
        authentication scheme should return `403 Permission Denied` responses.
        """
        pass

    def __call__(self, request):
        try:
            auth, user = self.authenticate(request)
        except AuthenticationError as exc:
            auth = AuthCredentials()
            user = UnauthenticatedUser(exc)
        finally:
            request.auth, request.user = auth, user
        return self.get_response(request)


class BasicAuthMiddleware(BaseAuthMiddleware):
    """
    HTTP Basic authentication against username/password.
    """

    www_authenticate_realm = 'api'

    def authenticate(self, request):
        """
        Returns a `User` if a correct username and password have been supplied
        using HTTP Basic authentication.  Otherwise returns `None`.
        """
        auth = get_authorization_header(request).split()

        if not auth or auth[0].lower() != b'basic':
            return None, None

        if len(auth) == 1:
            msg = _('Invalid basic header. No credentials provided.')
            raise AuthenticationError(msg)
        elif len(auth) > 2:
            msg = _('Invalid basic header. Credentials string should not contain spaces.')
            raise AuthenticationError(msg)
        try:
            auth_parts = base64.b64decode(auth[1]).decode().partition(':')
        except (TypeError, UnicodeDecodeError, binascii.Error):
            msg = _('Invalid basic header. Credentials not correctly base64 encoded.')
            raise AuthenticationError(msg)

        userid, password = auth_parts[0], auth_parts[2]
        return self.authenticate_credentials(userid, password, request)

    def authenticate_credentials(self, userid, password, request=None):
        """
        Authenticate the userid and password against username and password
        with optional request for context.
        """
        credentials = {get_user_model().USERNAME_FIELD: userid, 'password': password}
        user = authenticate(request=request, **credentials)

        if user is None:
            raise AuthenticationError(_('Invalid username/password.'))

        if not user.is_active:
            raise AuthenticationError(_('User inactive or deleted.'))

        return AuthCredentials(['authenticated']), user

    def authenticate_header(self, request):
        return 'Basic realm="%s"' % self.www_authenticate_realm



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['djgql', 'djgql.auth']

package_data = \
{'': ['*']}

install_requires = \
['django>=3,<4', 'python-gql<0.1']

setup_kwargs = {
    'name': 'django-gql',
    'version': '0.0.1',
    'description': '',
    'long_description': '# Django Graphql\n\n## gqlgen\n\ngqlgen is a generator tool for GraphQL.\n\n```shell script\nUsage: gqlgen [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  all             Generate all schema types\n  field-resolver  Generate field resolver.\n  type            Generate one type\n  type-resolver   Generate all schema types\n\n```\n\n\n## How to use\n\n```python\n# urls.py\nfrom django.contrib import admin\nfrom django.urls import path\n\nfrom djgql.views import GraphQLView\n\nurlpatterns = [\n    path(\'admin/\', admin.site.urls),\n    path(\'graphql/\', GraphQLView.as_view())\n]\n```\n\n```python\n# utils.py\ndef context_builder():\n    return {\'version\': 1}\n```\n\n```python\n# settings.py\nMIDDLEWARE = [\n    \'django.middleware.security.SecurityMiddleware\',\n    \'django.contrib.sessions.middleware.SessionMiddleware\',\n    \'django.middleware.common.CommonMiddleware\',\n    \'django.middleware.csrf.CsrfViewMiddleware\',\n    \'django.contrib.auth.middleware.AuthenticationMiddleware\',\n    \'django.contrib.messages.middleware.MessageMiddleware\',\n    \'django.middleware.clickjacking.XFrameOptionsMiddleware\',\n    \'djgql.auth.middleware.BasicAuthMiddleware\',\n]\nGRAPHQL_SCHEMA_FILE = os.path.join(BASE_DIR, \'starwar.gql\')\nGRAPHQL = {\n    \'SCHEMA\': \'starwar.schema.schema\',\n    \'ENABLE_PLAYGROUND\': True,\n    \'CONTEXT_BUILDER\': \'starwar.utils.context_builder\n}\n```\n\n```python\nimport typing\nfrom enum import Enum\n\nfrom django.conf import settings\nfrom gql import query, gql, type_resolver, enum_type, field_resolver\nfrom gql.build_schema import build_schema_from_file\nfrom djgql.auth import login_required\nfrom pydantic import BaseModel\n\ntype_defs = gql("""\ntype Query {\n    hello(name: String!): String!\n}\n""")\n\n\n@enum_type\nclass Episode(Enum):\n    NEWHOPE = 1\n    EMPIRE = 2\n    JEDI = 3\n\n\nclass Character(BaseModel):\n    id: typing.Text\n    name: typing.Optional[typing.Text]\n    friends: typing.Optional[typing.List[typing.Optional[\'Character\']]]\n    appears_in: typing.Optional[typing.List[typing.Optional[Episode]]]\n\n\nclass Human(Character):\n    id: typing.Text\n    name: typing.Optional[typing.Text]\n    friends: typing.Optional[typing.List[typing.Optional[Character]]]\n    appears_in: typing.Optional[typing.List[typing.Optional[Episode]]]\n    home_planet: typing.Optional[typing.Text]\n\n\nclass Droid(Character):\n    id: typing.Text\n    name: typing.Optional[typing.Text]\n    friends: typing.Optional[typing.List[typing.Optional[Character]]]\n    appears_in: typing.Optional[typing.List[typing.Optional[Episode]]]\n    primary_function: typing.Optional[typing.Text]\n\n\n@query\n@login_required\ndef hero(parent, info, episode: typing.Optional[Episode]) -> typing.Optional[Character]:\n    request = info.context[\'request\']\n    print(request.user)\n    return Human(id=\'test\')\n\n\n@field_resolver(\'Human\', \'name\')\ndef human_name(parent, info):\n    return \'Jack\'\n\n\n@type_resolver(\'Character\')\ndef resolve_character_type(obj, info, type_):\n    if isinstance(obj, Human):\n        return \'Human\'\n    if isinstance(obj, Droid):\n        return \'Droid\'\n    return None\n\n\nschema = build_schema_from_file(settings.GRAPHQL_SCHEMA_FILE)\n```',
    'author': 'syfun',
    'author_email': 'sunyu418@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/syfun/django-graphql',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



from typing import Any

from django.utils.translation import ugettext_lazy as _
from graphql import GraphQLError


class GraphQLExtensionError(GraphQLError):
    code = 'APOLLO_ERROR'
    message = 'apollo error'

    def __init__(self, message: str = None, **kwargs: Any):
        message = message or self.message
        extensions = {'code': self.code}
        if kwargs:
            extensions['exception'] = kwargs
        super().__init__(message=message, extensions=extensions)


class AuthenticationError(GraphQLExtensionError):
    code = 'AUTHENTICATION_ERROR'
    message = _('authentication error')


class ForbiddenError(GraphQLExtensionError):
    code = 'FORBIDDEN_ERROR'
    message = _('forbidden error')


class UserInputError(GraphQLExtensionError):
    code = 'USER_INPUT_ERROR'
    message = _('user input error')


class MethodNotAllowedError(GraphQLExtensionError):
    code = 'METHOD_NOT_ALLOWED'
    message = _('method not allowed, only accept ["GET", "POST"]')
